const io = @import("io.zig");

///
/// Number formatting modes supported by [writeInt].
///
pub const Radix = enum {
    binary,
    tinary,
    quaternary,
    quinary,
    senary,
    septenary,
    octal,
    nonary,
    decimal,
    undecimal,
    duodecimal,
    tridecimal,
    tetradecimal,
    pentadecimal,
    hexadecimal,
};

///
/// Writes `value` as a ASCII / UTF-8 encoded integer to `writer`, returning `true` if the full
/// sequence was successfully written, otherwise `false`.
///
/// The `radix` argument identifies which base system to format `value` as.
///
pub fn printInt(writer: io.Writer, radix: Radix, value: anytype) bool {
    const Int = @TypeOf(value);
    const type_info = @typeInfo(Int);

    switch (type_info) {
        .Int => {
            if (value == 0) return writer.writeByte('0');

            // TODO: Unhardcode this as it will break with large ints.
            var buffer = std.mem.zeroes([28]u8);
            var buffer_count = @as(usize, 0);
            var n1 = value;

            if ((type_info.Int.signedness == .signed) and (value < 0)) {
                // Negative value.
                n1 = -value;
                buffer[0] = '-';
                buffer_count += 1;
            }

            while (n1 != 0) {
                const base = @enumToInt(radix);

                buffer[buffer_count] = @intCast(u8, (n1 % base) + '0');
                n1 = (n1 / base);
                buffer_count += 1;
            }

            for (buffer[0 .. (buffer_count / 2)]) |_, i|
                std.mem.swap(u8, &buffer[i], &buffer[buffer_count - i - 1]);

            return (writer.call(.{buffer[0 .. buffer_count]}) == buffer_count);
        },

        // Cast comptime int into known-size integer and try again.
        .ComptimeInt => return writer.
            writeInt(radix, @intCast(std.math.IntFittingRange(value, value), value)),

        else => @compileError("value must be of type int"),
    }
}
